package lightcontainer.domains.client;

import lightcontainer.interfaces.ClientHandlerFFE;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;
import lightcontainer.protocol.rules.reader.SigninRule;
import lightcontainer.protocol.rules.writer.SignErrorRule;
import lightcontainer.protocol.rules.writer.SignOkRule;

import java.io.*;
import java.net.Socket;
import java.nio.charset.StandardCharsets;

/**
 * ClientHandler
 *
 * <!> UNICAST CLIENT <!>
 * Class communicating with the client, and
 * intercepting and sending files to the client.
 *
 * @version     1.1
 * @since       1.0
 *
 * @see         Runnable
 * @see         AutoCloseable
 * @author      Jérémi NIHART <j.nihart@student.helmo.be>
 */
public class ClientHandler implements Runnable, AutoCloseable {
    // Variables
    private ClientHandlerFFE fileFrontEnd;
    private final Socket client;
    private ProtocolRepository protocolRep;
    private boolean client_run;

    /**
     * Login of client.
     * State : NULL if not connected and Login string if connected
     */
    private String login;

    private BufferedReader reader;
    private PrintWriter writer;
    private ProtocolWriter.ProtocolResult response;

    // Constructor
    public ClientHandler(Socket client, ClientHandlerFFE ffe, ProtocolRepository protocolRep) {
        this.fileFrontEnd = ffe;
        this.client = client;
        this.protocolRep = protocolRep;
        this.client_run = false;
        initClient();
    }

    /**
     * Initialise the Client's Reader and Writer.
     *
     * @since       1.0
     *
     * @see         BufferedReader
     * @see         PrintWriter
     */
    private void initClient() {
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                this.client.getInputStream(),
                StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                this.client.getOutputStream(),
                StandardCharsets.UTF_8
            ), true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread Function
     * Start the dialogue with the client.
     *
     * @since       1.0
     */
    @Override
    public void run() {
        this.client_run = true;
        while (this.client_run) {
            try {
                String command = this.reader.readLine();
                if (command != null) System.out.println("Client: " + command);

                ProtocolReader.ProtocolResult ruleResult = protocolRep.executeReader(command + "\r\n");

                if (isConnected()) {
                    ruleResult.read(
                            this.client.getInputStream()
                    );
                    ProtocolWriter.ProtocolResult writerCommand = ruleResult.getResultCommand();

                    if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND) {
                        fileFrontEnd.newCommand(
                                writerCommand,
                                getLogin()); // Envoie dans la file de tâche FileFrontEnd en attente d'un traitement d'un StorBackEnd

                        // Attend la fin de la réalisation de la tâche
                        waitTaskResponse();

                        writer.write(response.getCommand()); // Renvoye au client
                        writer.flush();
                        response.write(this.client.getOutputStream()); // Ecrit au client si nécessaire
                    } else {
                        writer.write(writerCommand.getCommand()); // Renvoye au client
                        writer.flush();
                    }

                } else {
                    authentication(ruleResult);
                }
            } catch (IOException ignore) { }

        }
    }

    /**
     * Permet au Client d'attendre la fin de la réalisation de sa tâche
     */
    private void waitTaskResponse() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) { e.printStackTrace(); }
        }
    }

    public void respond(ProtocolWriter.ProtocolResult response) {
        this.response = response;
        synchronized (this) {
            this.notify();
        }
    }

    private void authentication(ProtocolReader.ProtocolResult ruleResult) {
        try {
            SigninRule.Result signinResult = (SigninRule.Result) ruleResult;
            if (signinResult.checkCredentials()) {
                this.login = signinResult.getLogin();
                ProtocolWriter.ProtocolResult signokResult = protocolRep.executeWriter(SignOkRule.NAME);
                writer.write(signokResult.getCommand());
                writer.flush();
                return;
            }
        } catch (ClassCastException ignored) {}

        ProtocolWriter.ProtocolResult signErrorResult = protocolRep.executeWriter(SignErrorRule.NAME);
        writer.write(signErrorResult.getCommand()); // Envoie SignError car echec de la connection
        writer.flush();
        this.close(); // Fermeture de la connection
    }

    /**
     * Permet de savoir si l'utilisateur s'est connecté (Avec login et mdp)
     * @return
     */
    private boolean isConnected() {
        return login != null;
    }

    /**
     * AutoClosable Function
     * Close the Client thread and resources.
     *
     * @since       1.0
     */
    @Override
    public void close() {
        if (this.client_run) {
            try {
                this.client_run = false;
                this.client.close();
            } catch (IOException ignored) { }
        }
    }

    public String getLogin() {
        return this.login;
    }
}
